package actionPilot;

import java.util.LinkedList;

import kinematics.Driver;
import kinematics.Dynamics;
import kinematics.DynamicsContainer;
import kinematics.NewPose;
import kinematics.Pose2D;
import kinematics.Timing;

/**
 * Pilot for holonomic platforms that works with {@link Move Moves}.
 * <p>
 * The pilot can combine up to three different moves into a single action.
 * Combined moves have to act on different dimensions (x, y or z) though.<br>
 * The pilot will queue actions. A new action is only started when the current
 * action has finished. See {@link #addAction}. <br>
 * The pilot is able to perform a smooth transition between two subsequent
 * actions. This is called action blending and will be active when blendMode is
 * on. <br>
 * The pilot calculates the optimum speed of the robot for each action. If
 * possible it will use maximum speed (as is set when creating the pilot). It
 * will also make sure that combined Moves end at the same time. If needed the
 * pilot lowers the speed of some of the Moves to achieve this.
 * 
 * @author Aswin
 * 
 */
public class ActionPilot {
  public static int          VERBOSE_ACTION = 1;
  public static int          VERBOSE_SPEED  = 2;
  public static int          VERBOSE_LOOP   = 4;
  public static int          VERBOSE_TIME   = 8;

  // actions
  private Move               rest           = new Rest(true, true, true);
  private LinkedList<Action> actions        = new LinkedList<Action>();
  private Action             action;
  private Action             nextAction;
  private Action             defaultAction  = new Action();

  // action related dynamics
  private Dynamics           maxDynamics;

  // action related timing
  private int                inactive       = 0;

  // physical model of the robot;
  private Driver             chassis;

  // current location of the robot;
  Pose2D                     currentPose;

  // configuration fields
  private boolean            blendMode      = false;
  private int                verbose        = 0;

  // fields for reusing objects
  private NewPose            speed          = new NewPose();
  private NewPose            speed2         = new NewPose();
  float                      twoPI          = (float) (Math.PI * 2);
  float                      PI             = (float) (Math.PI);

  private Heartbeat          heartbeat;
  private boolean            suspended;

  private Timing             actionTimer    = new Timing();

  // transition related fields
  private boolean            inTransition   = false;
  private float              transitionTime;
  private float              transitionStart;

  /**
   * Constructor for the ActionPilot
   * 
   * @param chassis
   *          The {@link Driver} object representing the robot and its motors
   * @param currentPose
   *          A {@link Pose2D} object holding the actual pose of the robot
   * @param maxDynamics
   *          A {@link Dynamics} object holding the maximum speed and
   *          acceleration values
   */
  public ActionPilot(Driver chassis, Pose2D currentPose, Dynamics maxDynamics) {
    this.chassis = chassis;
    this.currentPose = currentPose;
    this.maxDynamics = maxDynamics;
    activateNext();
    heartbeat = new Heartbeat();
    heartbeat.setDaemon(true);
    heartbeat.start();
  }

  /**
   * Returns a description of the action the pilot is executing
   * 
   * @return
   */
  public String getCurrentAction() {
    return action.toString();
  }

  /**
   * Indicates whether action blending is on;
   * 
   * @return true if blending is enabled
   */
  public boolean isBlendMode() {
    return blendMode;
  }

  /**
   * Enables or disables action blending. Action blending is the technique that
   * creates smooth transitions from one action to another. With action blending
   * enabled the robot always accelerates and decelerates within the limits
   * given by the Dynamics parameter in the pilots constructor.
   * 
   * @param blendMode
   *          true to enables action blending
   */
  public void setBlendMode(boolean blendMode) {

    this.blendMode = blendMode;
    if (blendMode) {
      prepareNextAction();
    }
  }

  /**
   * Instructs the pilot to provide debugging information
   * 
   * @param verbose
   */
  public void enableVerbose(int verbose_level) {
    verbose = verbose | verbose_level;
  }

  /**
   * Instructs the pilot not to provide debugging information
   * 
   * @param verbose
   */
  public void disableVerbose() {
    verbose = 0;
  }

  private boolean verbose(int verbose_level) {
    if ((verbose & verbose_level) == 0)
      return false;
    return true;
  }

  /**
   * Prevents the motors from running.
   */
  public void suspendMotorCommands() {
    suspended = true;
  }

  /**
   * Reactivates suspended motors.
   */
  public void resumeMotorCommands() {
    suspended = false;
  }

  /**
   * Adds an action to the pilots action queue.<br>
   * 
   * @param moves
   *          An action consist of one to three Move objects
   * @return False when to pilot cannot combine the Moves into a single action
   */
  public boolean addAction(Move... moves) {
    Action newAction = new Action();
    for (int i = 0; i < moves.length; i++) {
      if (!newAction.addMove(moves[i]))
        return false;
    }
    actions.add(newAction);
    if (verbose(VERBOSE_ACTION))
      System.out.println("Adding to action stack: " + newAction.toString());
    if (actions.size() == 1 && blendMode) {
      prepareNextAction();
    }
    return true;
  }

  /**
   * Clears all queued actions without affecting the current action
   */
  public void clearFutureActions() {
    actions.clear();
  }

  /**
   * Clears future actions and replaces the current action with the action
   * specified
   * 
   * @param lAction
   *          location (x,y) related action
   * @param hAction
   *          heading related action
   */
  public void immediateAction(Move... components) {
    clearFutureActions();
    addAction(components);
    activateNext();
  }

  /**
   * Initiates the next action to perform. If there are no actions in the queue
   * the default rest action will be initiated.
   */
  protected synchronized void activateNext() {
    inTransition = false;
    if (actions.isEmpty())
      action = defaultAction;
    else
      action = actions.pop();
    if (verbose(VERBOSE_ACTION)) {
      System.out.println("Activating action " + action.toString());
    }
    if (!blendMode) {
      action.setMaximumDynamics(maxDynamics);
      actionTimer.start();
    }
    else {
      prepareNextAction();
      actionTimer.start(transitionTime / 2);
    }
  }

  private synchronized void prepareNextAction() {
    if (actions.isEmpty())
      nextAction = defaultAction;
    else
      nextAction = actions.getFirst();
    if (verbose(VERBOSE_ACTION)) {
      System.out.println("Preparing next action " + nextAction.toString());
    }
    nextAction.setMaximumDynamics(maxDynamics);
  }

  /**
   * Puts the robot in motionless state immediately
   */
  public void rest() {
    clearFutureActions();
    activateNext();
  }

  /**
   * Gets the next action from the action queue without removing it from the
   * queue. If there are no actions in the queue the default rest action will be
   * returned.
   * 
   * @return The next composite action
   */
  protected Action getNext() {
    if (actions.isEmpty())
      return defaultAction;
    return actions.peek();
  }

  /**
   * Iteration element of the pilot.
   */
  protected void go() {
    if (blendMode)
      goBlend();
    else
      goNoBlend();
  }

  /**
   * Performs an action when action blending is enabled
   */
  private synchronized void goBlend() {
    float dt = actionTimer.elapsed();
    // Activate next action when current action is finished
    if ((inTransition && (transitionStart + transitionTime <= dt))) {
      if (verbose(VERBOSE_ACTION))
        System.out.println("Action ended.");
      activateNext();
      dt = actionTimer.elapsed();
    }
    if (verbose(VERBOSE_TIME)) {
      System.out.println("dt: " + dt);
    }
    if (!inTransition) {
      // Start transition to next action when cirecumstanses are right
      if (isDoingNothing()) {
        if (verbose(VERBOSE_ACTION))
          System.out.println("Action will blend, action is doing nothing.");
        inTransition = true;
        transitionTime = nextAction.accelerationTime;
        transitionStart = dt;
      }
      if (action.actionTime==Float.POSITIVE_INFINITY && nextAction != defaultAction) {
        if (verbose(VERBOSE_ACTION))
          System.out.println("Infinite action will blend, new action is available.");
        inTransition = true;
        transitionTime = nextAction.accelerationTime;
        transitionStart = dt;
      }


      action.getSpeed(speed, dt);

      if (dt >= action.actionTime - action.accelerationTime * .5f) {
        float t = nextAction.timeToBlend(speed);
        if (verbose(VERBOSE_TIME)) {
          System.out.println("blend time: " + t);
          System.out.println("time left: " + (action.actionTime - dt));
        }
        if (t * .5f >= action.actionTime - dt) {
          if (verbose(VERBOSE_ACTION))
            System.out.println("Action will blend, end off current action is near.");
          inTransition = true;
          transitionTime = t;
          transitionStart = dt;
        }

        if (inTransition && verbose(VERBOSE_TIME)) {
          System.out.println("Transition time: " + transitionTime);
          System.out.println("Transition start: " + dt);
        }
      }
    }
    else {
      float inTransitionTime = dt - transitionStart;

      action.getSpeed(speed, transitionStart + inTransitionTime * .5f);
      nextAction.getSpeed(speed2, inTransitionTime * .5f);
      float weight = inTransitionTime / transitionTime;
      if (verbose(VERBOSE_SPEED)) {
        System.out.println("Speed current action: " + speed.toString());
        System.out.println("Speed next action: " + speed2.toString());
        System.out.println("Weight: " + weight);
      }
      speed2.times(weight);
      speed.times(1 - weight);
      speed.plus(speed2);
    }
    if (verbose(VERBOSE_SPEED)) {
      System.out.println("Speed in global frame: " + speed.toString());
    }
    speed.rotate(-currentPose.getAngle());
    if (verbose(VERBOSE_SPEED)) {
      System.out.println("Speed in local frame: " + speed.toString());
    }
    if (!suspended)
      chassis.move(speed);
  }

  /**
   * Performs an action when action blending is disabled
   */
  private synchronized void goNoBlend() {
    if (actionTimer.elapsed() > action.actionTime) {
      if (verbose(VERBOSE_ACTION))
        System.out.println("Action finished, action time elapsed");
      activateNext();
    }
    if (isDoingNothing()) {
      if (verbose(VERBOSE_ACTION))
        System.out.println("Action finished, action is doing nothing");
      activateNext();
    }
    if (action.actionTime==Float.POSITIVE_INFINITY && !actions.isEmpty()) {
      if (verbose(VERBOSE_ACTION))
        System.out.println("Infinite action finished, new action is available");
      activateNext();
    }
    speed.set(0, 0, 0);
    action.getSpeed(speed, actionTimer.elapsed());
    if (verbose(VERBOSE_SPEED)) {
      System.out.println("Speed in global frame: " + speed.toString());
    }
    speed.rotate(-currentPose.getAngle());
    if (verbose(VERBOSE_SPEED)) {
      System.out.println("Speed in local frame: " + speed.toString());
    }
    if (!suspended)
      chassis.move(speed);
  }

  /**
   * Some actions run until a target has been reached. Upon reaching the target
   * the robot will remain motionless. This method detect such a motionless
   * state.
   * 
   * @return
   */
  private boolean isDoingNothing() {
    if ((action != defaultAction) && speed.isVeryCloseTo(Pose2D.ORIGIN))
      inactive++;
    else
      inactive = 0;
    if (inactive > 20)
      return true;
    else
      return false;
  }

  /**
   * Indicates wheather the robot is performing an action
   * 
   * @return
   */
  public boolean isBusy() {
    if (action == defaultAction && actions.isEmpty())
      return false;
    return true;
  }

  /**
   * This class starts governs the iteration of the pilot
   * 
   * @author Aswin
   * 
   */
  private class Heartbeat extends Thread {
    private Timing  timer;
    private float   currentInterval = 0;
    private float   minimumInterval = 0.03f;
    private float   maximumInterval = 0.1f;
    private NewPose acceleration    = new NewPose();
    private float[] targetDelta     = { 10, 10, 0.157f };

    public void run() {
      setPriority(Thread.MAX_PRIORITY);
      this.timer = new Timing(maximumInterval);
      while (true) {
        currentInterval = timer.delay();
        go();
        timer.setInterval(this.getNewInterval());
      }
    }

    /**
     * Interval optimalization routine for the pilot
     * 
     * @return The optimum interval in msec
     */
    private float getNewInterval() {
      float v;
      float targetInterval = maximumInterval;
      if (verbose(VERBOSE_LOOP)) {
        System.out.println("Actual loop time: " + currentInterval);
      }
      acceleration.minus(speed);
      acceleration.devide(currentInterval);
      for (int i = 0; i < 3; i++) {
        v = targetDelta[i] / Math.abs(acceleration.get(i));
        if (targetInterval > v)
          targetInterval = v;
      }
      if (targetInterval < minimumInterval)
        targetInterval = minimumInterval;
      acceleration.set(speed);
      if (verbose(VERBOSE_LOOP)) {
        System.out.println("Proposed loop time: " + targetInterval);
      }
      return targetInterval;
    }

  }

  /**
   * A wrapper to combine moves into an Action
   * 
   * @author Aswin
   * 
   */
  protected class Action {

    private Move[]   moves    = { rest, rest, rest };
    private float    actionTime;
    private float    accelerationTime;
    private NewPose  accel;
    private NewPose  initialSpeed;
    private Dynamics dynamics = new DynamicsContainer();

    /**
     * Sets the speed and acceleration for the action. From maximum dynamics and
     * other parameters actual speed and acceleration are calculated and passed
     * on to the moves within the action.
     * 
     * @param max
     */
    public void setMaximumDynamics(Dynamics max) {
      // set maximum speeds and calculate actionTime;
      // TODO: lower maximum speed for very short actions
      float maxTime = 0;
      float[] actionTimes = new float[3];
      for (int d = 0; d < 3; d++) {
        moves[d].setDynamics(max, currentPose);
        actionTimes[d] = moves[d].getActionTime();
        if (actionTimes[d] != Float.POSITIVE_INFINITY)
          maxTime = Math.max(maxTime, actionTimes[d]);
      }
      if (maxTime == 0)
        actionTime = Float.POSITIVE_INFINITY;
      else
        actionTime = maxTime;
      // adjust speeds (all non Infinity actions should last as long as the
      // longest action);
      if (actionTimes[0] == Float.POSITIVE_INFINITY || actionTimes[0] == Float.POSITIVE_INFINITY) {
        dynamics.setLinearSpeed(max.getLinearSpeed());
      }
      else {
        dynamics.setLinearSpeed(max.getLinearSpeed() * Math.min(actionTimes[0], actionTimes[1]) / maxTime);
      }
      if (actionTimes[2] == Float.POSITIVE_INFINITY) {
        dynamics.setAngularSpeed(max.getAngularSpeed());
      }
      else {
        dynamics.setAngularSpeed(max.getAngularSpeed() * actionTimes[2] / maxTime);
      }
      // adjust acceleration(linear and angular acceleration should last take
      // the same time to go to full speed);
      accelerationTime = Math.max(dynamics.getAngularSpeed() / max.getAngularAcceleration(), dynamics.getLinearSpeed() / max.getLinearAcceleration());
      dynamics.setLinearAcceleration(dynamics.getLinearSpeed() / accelerationTime);
      dynamics.setAngularAcceleration(dynamics.getAngularSpeed() / accelerationTime);
      for (int d = 0; d < 3; d++) {
        moves[d].setDynamics(dynamics, currentPose);
      }
      if (verbose(VERBOSE_TIME)) {
        System.out.println("Action time: " + actionTime);
        System.out.println("Accel time: " + accelerationTime);
      }
    }

    /**
     * Calculates the speed over the three dimensions at a given moment
     * 
     * @param speed
     *          The {@link Pose2D} speed object to hold the speed.
     * @param dt
     *          The moment in time expressed in seconds elapsed since the start
     *          of the action
     */
    public void getSpeed(NewPose speed, float dt) {
      for (int d = 0; d < 3; d++) {
        moves[d].setTime(dt);
        speed.set(d, moves[d].get(d));
      }
    }

    /**
     * Add moves to the newly created Action. Move may not have overlapping
     * dimensions. Otherwise the method returns a false condition.
     * 
     * @param move
     * @return
     */
    protected boolean addMove(Move move) {
      for (int d = 0; d < 3; d++) {
        if (move.supportsDimension(d)) {
          if (moves[d] != rest)
            return false;
          moves[d] = move;
        }
      }
      return true;
    }

    @Override
    public String toString() {
      return moves[0].toString() + "," + moves[1].toString() + "," + moves[2].toString();
    }

    /** Calculates the time needed to go from current speed to the initial speed of this action
     * @param currentSpeed
     * @return
     */
    protected float timeToBlend(Pose2D currentSpeed) {
      if (initialSpeed == null) {
        initialSpeed = new NewPose();
        accel = new NewPose();
        getSpeed(initialSpeed, 0);
      }
      accel.set(initialSpeed);
      accel.minus(currentSpeed);
      return Math.max(accel.length() / dynamics.getLinearAcceleration(), Math.abs(accel.getAngle()) / dynamics.getAngularAcceleration());

    }

  }

}
