package actionPilot;

import kinematics.NewPose;

/**
 * This class is the base for location based moves. <br>
 * Location based moves have some external reference point that they want to go to. 
 * The x,y and z/angle fields of these moves represent distance to target. 
 * <p>This base move has the getters for x,y and z modified to convert distance to speed using current dynamics. 
 * Subclasses of this class therefore do not have to bother with converting distance to speed.
 * 
 * @author Aswin
 * 
 */
public abstract class LocationBasedMove extends Rest {
  boolean verbose = false;

  public LocationBasedMove(boolean restX, boolean restY, boolean restAngle) {
    super(restX, restY, restAngle);
  }

  @Override
  public float getX() {
    double maxSpeed = 0;
    if (x != 0) {
      double dist = Math.sqrt(x * x + y * y);
      if (dist >= NewPose.CLOSE_DISTANCE) {
        double direction = Math.atan2(y, x);
        double remweg = (Math.pow(dynamics.getLinearSpeed(), 2) / (2 * dynamics.getLinearAcceleration()));
        if (dist >= remweg)
          maxSpeed =  dynamics.getLinearSpeed();
        else
          maxSpeed =  Math.sqrt(dist * 2 * dynamics.getLinearAcceleration());
        maxSpeed = (maxSpeed * Math.cos(direction));
      }
    }
    if (verbose)
      System.out.println("x: " + x + " -> " + maxSpeed);
    return (float) maxSpeed;
  }

  @Override
  public float getY() {
    double maxSpeed = 0;
    if (x != 0) {
      double dist = Math.sqrt(x * x + y * y);
      if (dist >= NewPose.CLOSE_DISTANCE) {
        double direction = Math.atan2(y, x);
        double remweg = (Math.pow(dynamics.getLinearSpeed(), 2) / (2 * dynamics.getLinearAcceleration()));
        if (dist >= remweg)
          maxSpeed =  dynamics.getLinearSpeed();
        else
          maxSpeed =  Math.sqrt(dist * 2 * dynamics.getLinearAcceleration());
        maxSpeed = (maxSpeed * Math.sin(direction));
      }
    }
    if (verbose)
      System.out.println("y: " + y + " -> " + maxSpeed);
    return (float) maxSpeed;
  }

  @Override
  public float getZ() {
    double maxSpeed = 0;
    if (Math.abs(z) >= NewPose.CLOSE_ANGLE) {
      float sign = Math.signum(z);
      double remweg = (Math.pow(dynamics.getAngularSpeed(), 2) / (2 * dynamics.getAngularAcceleration()));
      if (z >= remweg)
        maxSpeed = sign * dynamics.getAngularSpeed();
      else
        maxSpeed = sign * (float) (Math.sqrt(Math.abs(z) * 2 * dynamics.getAngularAcceleration()));
    }
    if (verbose)
      System.out.println("z: " + z + " -> " + maxSpeed);
    return (float) maxSpeed;
  }

  @Override
  public float get(int i) {
    switch (i) {
      case 0:
        return getX();
      case 1:
        return getY();
      case 2:
        return getZ();
      default:
        throw new IndexOutOfBoundsException();
    }
  }

}
