package actionPilot;

import kinematics.Dynamics;
import kinematics.Pose2D;


/**
 * A Move represents a move that a robot can make. <p>
 * A move defines the robots speed at any moment in time since the start of the action.
 * Moves are combined into an action. A pilot executes the action.
 * @author Aswin
 *
 */
public interface Move {
	
	
	/** Sets the dynamics (speed and acceleration) that the move should use to calculate actionTime and speed at any moment
	 * @param dynamics
	 * @param currentPose
	 */
	public void setDynamics(Dynamics dynamics, Pose2D currentPose);
	
	
	/**
	 * Called by the pilot to set the time elapsed since the start of the action
	 * @param time
	 * @return
	 * false, if the action has ended, else true.
	 */
	public boolean setTime(float time);
	

	/**
	 * Returns the time (in seconds) that this movw needs to complete.
	 * @return
	 * time in seconds. Both an infinite lasting move and a move of which the timespan is unknown are indicated with Float.POSITIVE_INFINITY
	 */
	public float getActionTime();
	
	/** A move can occur in any combination of X, Y and Angle dimensions. 
	 * This function returns an indication of the move takes place in the specified dimension. <br>
	 * Most moves take place in both X and Y dimension, lines and arcs for example. Other moves taake place in the Angle dimension, like a rotation.
	 * 
	 * @param dimension
	 * The index number of the dimension, x=0, y=1, angle=2.
	 * @return
	 * true or false
	 */
	public boolean supportsDimension(int dimension) ;
	  
	/** Returns a true if the move alters the X position of the robot
	 * @return
	 */
	public boolean supportsX();
	
  /** Returns a true if the move alters the y position of the robot
   * @return
   */
  public boolean supportsY();

  /** Returns a true if the move alters the orientation of the robot
   * @return
   */
  public boolean supportsAngle();
  
  /** Returns the speed of the robot in the X dimension at any given time. The time has to be set beforehand using {@link #setTime(float)}.
   * @return
   */
  public float getX();
  
  /** Returns the speed of the robot in the Y dimension at any given time. The time has to be set beforehand using {@link #setTime(float)}.
   * @return
   */
  public float getY();
  
  /** Returns the speed of the robot in the Angle dimension at any given time. The time has to be set beforehand using {@link #setTime(float)}.
   * @return
   */
  public float getAngle();
  
  /** Returns the speed of the robot in the Nth dimension at any given time. The time has to be set beforehand using {@link #setTime(float)}.
   * @param index
   * The index number of the dimension, x=0, y=1, angle=2.
   * @return
   */
public float get(int index);

}
