package kinematics;


/** A Pose object to hold a pose in a two dimensional space. 
 * The heading or angle component of this pose is expressed in radians. <p>
 * See {@link Pose2D} for details.
 * @author Aswin
 *
 */
public class NewPose extends Vector implements Pose2D {
  protected static float PI             = (float) Math.PI;
  protected static float TWOPI          = (float) Math.PI * 2;
  protected static float CLOSE_DISTANCE = 4;
  protected static float CLOSE_ANGLE    = 4*PI /180;

  public NewPose(Pose2D target) {
    x = target.getX();
    y = target.getY();
    z = target.getZ();
  }

  public NewPose() {
    x = 0;
    y = 0;
    z = 0;
  }

  public NewPose(float x, float y, float angle) {
    this.x = x;
    this.y = y;
    this.z = angle;
  }

  
  @Override
  public float getAngle() {
    return ((z + PI) % TWOPI) - PI;
  }
  
  @Override
  public void setAngle(float angle) {
    z=angle;
  }

  @Override
  public int getRotation() {
    return (int) (z / TWOPI);
  }

  public void setPose(float x, float y, float h) {
    set(x, y, h);
  }

  /**
   * Defines the boundaries for close by test
   * 
   * @param distance
   *          distance between the two poses
   * @param angle
   *          angle between the two poses
   */
  public void setClose(float distance, float angle) {
    CLOSE_DISTANCE = distance;
    CLOSE_ANGLE = angle;
  }

  @Override
  public float distanceTo(Pose2D pose) {
    float deltaX = x - pose.getX();
    float deltaY = y - pose.getY();
    return (float) Math.sqrt(deltaX * deltaX + deltaY * deltaY);
  }

  @Override
  public float angleTo(Pose2D pose) {
    return pose.getAngle() - this.getAngle();
  }

  @Override
  public float directionTo(Pose2D pose) {
    return (float) Math.atan2(pose.getY() - y, pose.getX() - x);
  }

  @Override
  public float length() {
    return (float) Math.sqrt(x * x + y * y);
  }

  /**
   * Normalizes the length of the pose to 1 while maintaining the angle to the
   * origin. In other words, it sets the distance to the origin to one
   * respecting the ratio between the X and Y position of the pose
   * 
   * @return this
   */
  public Pose2D normalize() {
    return normalize(1);
  }

  /**
   * Normalizes the length of the pose to a given value while maintaining the
   * angle to the origin. In other words, it sets the distance to the origin to
   * the given value respecting the ratio between the X and Y position of the
   * pose
   * 
   * @param length
   *          The length to normilize to
   * @return this
   */
  public Pose2D normalize(float length) {
    float factor = length / this.length();
    x *= factor;
    y *= factor;
    return this;
  }

  public void rotate(float angle) {
    float xt, yt;
    float cos = (float) Math.cos(angle);
    float sin = (float) Math.sin(angle);
    xt = cos * x - sin * y;
    yt = sin * x + cos * y;
    x = xt;
    y = yt;
  }

  public boolean isVeryCloseTo(Pose2D pose) {
    // TODO: make boundary flexible
    if (this.distanceTo(pose) > CLOSE_DISTANCE)
      return false;
    if (Math.abs(getAngle() - pose.getAngle()) > CLOSE_ANGLE)
      return false;
    return true;
  }

  public void setPose(float x, float y, float angle, int rotations) {
    setPose(x, y, rotations * TWOPI + angle);
  }

  public String toString() {
    return "X: " + String.format("%7.3f",x) + " Y: " + String.format("%7.3f",y) + " A: " +  String.format("%7.3f",z) ;
  }
}
