package samples;

import actionPilot.*;
import kinematics.*;
import lejos.hardware.Brick;
import lejos.hardware.BrickFinder;
import lejos.hardware.Button;
import lejos.hardware.Sound;
import lejos.hardware.ev3.LocalEV3;
import lejos.hardware.motor.EV3LargeRegulatedMotor;
import lejos.hardware.port.MotorPort;
import lejos.hardware.sensor.EV3IRSensor;
import lejos.hardware.sensor.EV3TouchSensor;
import lejos.hardware.sensor.EV3UltrasonicSensor;
import lejos.hardware.sensor.MindsensorsAccelerometer;
import lejos.hardware.sensor.SensorMode;
import lejos.robotics.SampleProvider;
import lejos.robotics.filter.Dump;
import lejos.robotics.filter.MeanFilter;
import lejos.robotics.filter.OffsetCorrectionFilter;
import lejos.robotics.subsumption.Arbitrator;
import lejos.robotics.subsumption.Behavior;
import lejos.utility.Delay;

/**
 * Test program for the ActionPilot.
 * cd /home/lejos/programs/
 * jrun -cp Pilot.jar samples/Test
 * 
 * @author Aswin
 * 
 */
public class Rover {
  static float      PI      = (float) Math.PI;
  static float      TWOPI   = (float) Math.PI * 2;
  ActionPilot       pilot;
  Driver            robot;
  Odometer          odometer;
  float direction=0;
  private float[] rangeIRSample;
  private SensorMode rangeIR;
  private EV3IRSensor irSensor;
  private Arbitrator arbitrator;
  private EV3UltrasonicSensor usSensor;
  private SampleProvider rangeUS;
  private float[] rangeUSSample;
  private MindsensorsAccelerometer accSensor;
  private SampleProvider collisionDetector;
  private float[] collisionSample;

  /**
   * @param args
   */
  public static void main(String[] args) {
    new Rover();
  }

  private Rover() {
    buildDriveTrain();
    buildSensors();
    buildBehavior();
    Sound.beep();
    while (!Button.ENTER.isDown())
      Delay.msDelay(50);
    arbitrator.start();
    irSensor.close();
  }
  
  private void buildBehavior() {
    Behavior[] behaviorList = {  new Drive(), new LookAround(), new Collision(), new Stop()};
    arbitrator=new Arbitrator(behaviorList,true);
  }
  
  private void buildSensors() {
    Brick brick=LocalEV3.get();
    // IR sensor
    irSensor= new EV3IRSensor(brick.getPort("S2"));
    rangeIR=irSensor.getDistanceMode();
    rangeIRSample=new float[rangeIR.sampleSize()];
    
    // US sensor
    usSensor=new EV3UltrasonicSensor(brick.getPort("S3"));
    rangeUS = usSensor.getDistanceMode();
    rangeUSSample=new float[rangeUS.sampleSize()];
    
    // Accelerometer
    accSensor=new MindsensorsAccelerometer(brick.getPort("S4"));
    collisionDetector=new MeanFilter(new VectorLengthFilter(new ErrorTermFilter(accSensor)),3);
    collisionSample=new float[collisionDetector.sampleSize()];
  }
  
  protected float getRange() {
    rangeIR.fetchSample(rangeIRSample, 0);
    rangeUS.fetchSample(rangeUSSample, 0);
    return Math.min(rangeIRSample[0]/100, rangeUSSample[0]);
  }

  protected boolean getTouch() {
    return Button.ESCAPE.isDown() ? true : false;
  }
 
  private void buildDriveTrain() {
    Dynamics maxDynamics=new DynamicsContainer(400,400,PI/4,PI/2);
    ModelBuilder build = new ModelBuilder(3);
    float gear = - 24 / 12;
    build.inRadians();
    build.wheelPolarPosition(0, 136, PI, new EV3LargeRegulatedMotor(MotorPort.B), gear);
    build.wheelPolarPosition(1, 136, (PI * 1 / 3), new EV3LargeRegulatedMotor(MotorPort.A), gear);
    build.wheelPolarPosition(2, 136, (PI * 5 / 3), new EV3LargeRegulatedMotor(MotorPort.C), gear);
    robot = new Chassis(build);
    odometer = new Odometer(build);
    pilot = new ActionPilot(robot, odometer,maxDynamics);
    pilot.setBlendMode(true);
  }
  
  private class Drive implements Behavior {
    boolean pleaseStop=false;

    @Override
    public boolean takeControl() {
        return true;
    }

    @Override
    public void action() {
      pleaseStop=false;
      Move travel=new Travel(odometer.getAngle());
      Move rotateTo=new RotateTo(odometer.getAngle());
      pilot.addAction(travel,rotateTo);
      while (!pleaseStop) {
        Delay.msDelay(10);
      }
    }

    @Override
    public void suppress() {
      pleaseStop=true;
    }

  }
  
  private class LookAround implements Behavior {
    boolean pleaseStop=false;

    @Override
    public boolean takeControl() {
      if (getRange()<0.44) {
        return true;
      }
      return false;
    }
    
    
    @Override
    public void action() {
      pleaseStop=false;
    Move rotateTo=new RotateTo(odometer.getAngle()+(float)Math.signum(Math.random()-0.5)*PI);
    pilot.addAction(rotateTo);

      while (getRange()<0.65 && !pleaseStop) {
        Delay.msDelay(10);
      }
    }

    @Override
    public void suppress() {
      pleaseStop=true;
    }
    
  }
  
  private class Collision implements Behavior {
    boolean pleaseStop=false;

    @Override
    public boolean takeControl() {
      collisionDetector.fetchSample(collisionSample,0);
      //System.out.println(collisionSample[0]);
      if (collisionSample[0]>1.5f) {
        return true;
      }
      return false;
    }

    @Override
    public void action() {
      System.out.println("Start");
      pleaseStop=false;
      Move travel=new Line(50,odometer.getAngle()-PI);
      pilot.immediateAction(travel);
      while ((!pleaseStop) && pilot.isBusy()) {
        Delay.msDelay(40);
      }     
      System.out.println("End");

    }

    @Override
    public void suppress() {
      pleaseStop=true;
    }
    
  }
    
  private class Stop implements Behavior {

    @Override
    public boolean takeControl() {
      return getTouch() ?  true : false;
    }

    @Override
    public void action() {
      pilot.rest();
      while (pilot.isBusy()) {
        Delay.msDelay(50);
      }
      System.exit(0);
    }

    @Override
    public void suppress() {
    }
    
  }
}
