package kinematics;

import lejos.robotics.RegulatedMotor;

/**
 * The modelBuilder class builds a kinematic model or a reverse kinematic model
 * of a robot. The kinematic model can be used to translate movements of a robot
 * into movement of its wheels. The reverse kinematic model can be used to
 * translate wheel position into robot pose.
 * <p>
 * 
 * The modelBuilder class assumes the robot has a holonomic
 * drive system. A minimum of three motorized
 * wheels is required. 
 * <p>
 * 
 * The model builder allows one to describre the dimensions of the robots drive
 * system. The drive system consists of a number of wheels, each with its own
 * dimensions and location. A wheel can be connected to a motor via a gear
 * train.
 * <p>
 * 
 * Once the drive system is fully described the ModelBuilder can create a matrix
 * containing the forward or reverse kinematic model.
 * 
 * @author Aswin Bouwmeester
 * 
 */
public class ModelBuilder {
  // TODO: Make it work for degrees

  static int             HOLONOMIC    = 0;
  static int             DIFFERENTIAL = 1;
  static double          PI           = Math.PI;

  final int              numberOfWheels;
  final RegulatedMotor[] motor;
  final double[]         gearRatio;
  final double[]         wheelRadius;
  final double[]         wheelAngle;
  final double[]         wheelX;
  final double[]         wheelY;
  final int[]            ticks;
  boolean                degrees      = false;
  int                    platformType = HOLONOMIC;
  boolean                verbose      = false;

  /**
   * The constructor of the ModelBuilder. It needs the number of wheels of the
   * robot as input.
   * 
   * @param numberOfWheels
   *          number of (motorized) wheelsof the robot.
   */
  public ModelBuilder(int numberOfWheels) {
    this.numberOfWheels = numberOfWheels;
    motor = new RegulatedMotor[numberOfWheels];
    gearRatio = new double[numberOfWheels];
    wheelRadius = new double[numberOfWheels];
    wheelAngle = new double[numberOfWheels];
    wheelX = new double[numberOfWheels];
    wheelY = new double[numberOfWheels];
    ticks = new int[numberOfWheels];
    resetToDefault();
  }

  /**
   * Builds the forward kinematic model. This model has the form of a matrix
   * where each row represents a single wheel. The matrix is used to translate
   * movement, speed or acceleration of the robot into movement, speed or
   * acceleration of the wheels using matrix algebra. For example
   * <p>
   * <code>
   * Matrix wheelSpeed=kinematicModel.times(robotSpeed);
   * </code>
   * 
   * @return A matrix of the kinematic model of the robot.
   */
  public Matrix buildForwardKinematicModel() {
    Matrix k = buildKinematicModel();
    Matrix f = buildFactorMatrix();
    k.arrayTimesEquals(f);
    return k;
  }

  /**
   * Builds the reverse kinematic model. This model has the form of a matrix
   * where each column represents a single wheel. The matrix is used to
   * translate location, speed or acceleration of the wheels into location,
   * speed or acceleration of the robot using matrix algebra. For example
   * <p>
   * <code>
   * Matrix maxRobotSpeed=reverseKinematicModel.times(maxWheelSpeed);
   * </code>
   * 
   * @return A matrix of the reverse kinematic model of the robot.
   */
  public Matrix buildReverseKinematicModel() {
    int lastRow;
    Matrix k = buildForwardKinematicModel();
    if (platformType == HOLONOMIC)
      lastRow = 2;
    else
      lastRow = 1;
    if (k.getRowDimension() > lastRow)
      k = k.getMatrix(0, lastRow, 0, 2);
    return k.inverse();
  }

  /**
   * Builds an array of motors corresponding the the kinematic model;
   * 
   * @return
   */
  public RegulatedMotor[] buildMotorArray() {
    return motor;
  }

  private void dump(int i) {
    print("Wheel: " + i);
    print(" Location:");
    print("  Cartesian: " + wheelX[i] + ", " + wheelY[i]);
    print("  Polar: " + getDistance(i) + ", " + getTheta(i));
    print(" Direction:");
    print("  Angle: " + wheelAngle[i]);
    print("  Offset: " + (wheelAngle[i] - (Math.atan2(wheelY[i], wheelX[i]) - Math.PI / 2)));
    print(" Radius: " + wheelRadius[i]);
    print(" Gear ratio: " + gearRatio[i]);
    print(" Encoder ticks: " + ticks[i]);
    if (motor[i] != null)
      print(" Motor: " + motor[i].toString());
    print(" ");
  }

  private void print(String s) {
    System.out.println(s);
  }

  private double getDistance(int i) {
    return Math.sqrt(wheelX[i] * wheelX[i] + wheelY[i] * wheelY[i]);
  }

  private double getTheta(int i) {
    return Math.atan2(wheelY[i], wheelX[i]);
  }

  private Matrix buildKinematicModel() {
    Matrix m = new Matrix(numberOfWheels, 3);
    for (int i = 0; i < numberOfWheels; i++) {
      if (verbose)
        dump(i);
      m.set(i, 0, (float) Math.cos(wheelAngle[i]));
      m.set(i, 1, (float) -Math.sin(wheelAngle[i]));
      double dif = wheelAngle[i] - getTheta(i) - Math.PI / 2; 
      m.set(i, 2, (float) (-getDistance(i) * Math.cos(dif)));
    }
    return m;
  }

  private Matrix buildFactorMatrix() {
    Matrix m = new Matrix(numberOfWheels, 3);
    for (int i = 0; i < numberOfWheels; i++) {
      m.set(i, 0, (float) (ticks[i] / (gearRatio[i] * 2 * Math.PI * wheelRadius[i])));
      m.set(i, 1, (float) (ticks[i] / (gearRatio[i] * 2 * Math.PI * wheelRadius[i])));
      m.set(i, 2, (float) (ticks[i] / (gearRatio[i] * 2 * Math.PI * wheelRadius[i])));
    }
    return m;

  }

  private void resetToDefault() {
    for (int index = 0; index < numberOfWheels; index++) {
      this.motor[index] = null;
      this.gearRatio[index] = 1;
      this.wheelRadius[index] = 24;
      this.wheelAngle[index] = 0;
      this.wheelX[index] = 0;
      this.wheelY[index] = 0;
      this.ticks[index] = 360;
    }
  }

  /**
   * Instructs the ModelBuilder that angles are specified in degrees
   * 
   * @return this
   */
  public ModelBuilder inDegrees() {
    // TODO: Make it work
    degrees = true;
    return this;
  }

  /**
   * Instructs the ModelBuilder that angles are specified in radians
   * 
   * @return this
   */
  public ModelBuilder inRadians() {
    degrees = false;
    return this;
  }

  /**
   * Specifies the motor that is connected to the Nth wheel
   * 
   * @param index
   *          Index number of the wheel
   * @param motor
   *          a Regulated motor
   * @return this
   */
  public ModelBuilder motor(int index, RegulatedMotor motor) {
    this.motor[index] = motor;
    return this;
  }

  /**
   * Specifies the gear ratio of the geartrain of the Nth wheel
   * <P>
   * The gear ratio is the ratio between motor rotation and wheel rotation. A
   * ratio of three means the wheel is turning three times as fast as the motor.
   * Gear ratios can be negative to express that the direction of rotation is
   * reversed. Negative gear ratos can also be used to indicate a motor that is
   * mounted reversed. By default the gear ratio equals one.
   * 
   * @param index
   *          Index number of the wheel
   * @param gearRatio
   *          the ratio between motor rotation and wheel rotation
   * @return this
   */
  public ModelBuilder gearRatio(int index, float gearRatio) {
    this.gearRatio[index] = gearRatio;
    return this;
  }

  /**
   * Specifies the radius of the wheel in the unit one prefers.
   * 
   * @param index
   *          Index number of the wheel
   * @param wheelRadius
   *          size of the wheel from center to outer rim
   * @return this
   */
  public ModelBuilder wheelRadius(int index, float wheelRadius) {
    this.wheelRadius[index] = wheelRadius;
    return this;
  }

  /**
   * Specifies the diameter of the wheel in the unit one prefers.
   * 
   * @param index
   *          Index number of the wheel
   * @param wheelRadius
   *          diameter of the wheel
   * @return this
   */
  public ModelBuilder wheelDiameter(int index, float wheelDiameter) {
    this.wheelRadius[index] = wheelDiameter / 2;
    return this;
  }

  /**
   * Specifies the location of the wheel in respect to the origin of the robot
   * 
   * @param index
   *          Index number of the wheel
   * @param x
   *          The foward location of the wheel. A positive value indicates a
   *          wheel in front the origin of the robot
   * @param y
   *          The sideways location of the wheel. A positive value indicates a
   *          wheel to the left of the origin.
   * @param angle
   *          The angle between the robot x-axis and the forward pointing angle
   *          of the wheel
   * @return this
   */
  public ModelBuilder wheelCartesianPosition(int index, float x, float y, float angle) {
    wheelX[index] = x;
    wheelY[index] = y;
    wheelAngle[index] = angle;
    return this;
  }

  /**
   * Specifies the location of the wheel in respect to the origin of the robot
   * 
   * @param index
   *          Index number of the wheel
   * @param x
   *          The foward location of the wheel. A positive value indicates a
   *          wheel in front the origin of the robot
   * @param y
   *          The sideways location of the wheel. A positive value indicates a
   *          wheel to the left of the origin.
   * @param angle
   *          The angle between the robot x-axis and the forward pointing angle
   *          of the wheel
   * @param motor
   *          a Regulated motor
   * @param gearRatio
   *          the ratio between motor rotation and wheel rotation
   * @return this
   */  public ModelBuilder wheelCartesianPosition(int index, float x, float y, float angle, RegulatedMotor motor, double gearRatio) {
    wheelCartesianPosition(index, x, y, angle);
    this.motor[index] = motor;
    this.gearRatio[index] = gearRatio;
    return this;
  }

  /**
   * Specifies the location and orientation of the wheel in respect to the
   * origin of the robot using polar location. The method assumes the wheel axis
   * goes throught the origin of the robot.
   * 
   * @param index
   *          Index number of the wheel
   * @param radius
   *          The radius odf the polar position, ie the distance between the wheel and origin of the robot.
   * @param angle
   *          The angle of hte polar position, ie the angle between wheel axle and x axis of the robot.
   * @return this
   */
  public ModelBuilder wheelPolarPosition(int index, float radius, float angle) {
    wheelX[index] = (float) (Math.cos(angle) * radius);
    wheelY[index] = (float) (Math.sin(angle) * radius);
    this.wheelAngle[index] = angle - PI / 2;
    return this;
  }

  /**
   * Specifies the location and orientation of the wheel in respect to the
   * origin of the robot using polar location. The method assumes the wheel axis
   * goes throught the origin of the robot.
   * 
   * @param index
   *          Index number of the wheel
   * @param distanceToOrigin
   *          The distance between the wheel and origin of the robot.
   * @param angle
   *          The angle between the x-axis of the robot and the wheel axle.
   * @param motor
   *          a Regulated motor
   * @param gearRatio
   *          the ratio between motor rotation and wheel rotation
   * @return this
   */
  public ModelBuilder wheelPolarPosition(int index, float distanceToOrigin, float angle, RegulatedMotor motor, double gearRatio) {
    wheelPolarPosition(index, distanceToOrigin, angle);
    this.motor[index] = motor;
    this.gearRatio[index] = gearRatio;
    return this;
  }

  /**
   * Specifies the number of encoder ticks for one full rotation of the motor.
   * Use this method only for non NXT motors.
   * 
   * @param index
   *          Index number of the wheel
   * @param ticks
   *          number of encoder ticks that corresponds to one full rotaion of
   *          the motor.
   * @return
   */
  public ModelBuilder encodersPerRotation(int index, int ticks) {
    this.ticks[index] = ticks;
    return this;
  }

}
