package kinematics;


/**
 * Represents the position and orientation of an object in a two dimensional
 * space.<p>
 * The position is a cartesian position (x,y coordinate pair) in respect
 * to a reference.<br>
 * The angle is the orientation between the XY-frame of the
 * object and the XY-frame of the reference expresssed in radians. 
 * A positive angle indicates a counter clockwise rotation of the pose.
 * In addition to the angle a the Pose2D also knows the rotation. This is the
 * number of full rotations the object has made in respect to the reference
 * frame.<p>
 * 
 * @author Aswin
 * 
 */
public interface Pose2D extends Vector3D {

  Pose2D ORIGIN = new NewPose();

  /**
   * Returns the angle between the object and the reference frame
   * 
   * @return angle expressed in radians
   */
  public float getAngle();

  /**
   * Returns the number of full rotations the object has made 
   * 
   * @return the number of full rotations
   */
  public int getRotation();

  /**
   * Returns the distance between this pose and another pose.
   * 
   * @param pose
   * @return the distance expressed in the unit that is used for the Poses
   */
  public float distanceTo(Pose2D pose);
  
  /** 
   * Returns the angle between this pose and the given pose 
   * 
   * @param pose
   * @return
   * the angle in radians
   */
  public float angleTo(Pose2D pose);

  /**
   * Returns the distance between this pose and the origin of the reference
   * frame
   * 
   * @return the length expressed in the unit that is used for the Pose
   */
  public float length();

  /** Returns the direction of the vector going from this pose to the specified pose
   * @param pose
   * @return
   * direction in radians
   */
  public float directionTo(Pose2D pose);

  public void setAngle(float angle);


}